/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/bincimapd.cc
 *  
 *  Description:
 *    Implementation of the main bincimapd service
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#include <string>

#include <signal.h>

#include "broker.h"
#include "depot.h"
#include "exceptions.h"
#include "io/io.h"
#include "io/io-ssl.h"
#include "mailbox/maildir/maildir.h"
#include "operators.h"
#include "session.h"
#include "bincimapd-module.h"

using namespace ::std;
using namespace Binc;

namespace Binc {
  void initConfig(void);
}

vector<BincModule> modules;

namespace {
  void alarmhandler(int signum)
  {
    IO &com = IOFactory::getInstance().get(1);
    IO &logger = IOFactory::getInstance().get(2);

    logger << "Server closed connection (no activity from client)" << endl;
    com << "* BYE Binc IMAP shutting down - timeout exceeded" << endl;
    com.flushContent();
    logger.flushContent();
    exit(0);
  }

  //------------------------------------------------------------------------
  void usage(char *name)
  {
    IO &logger = IOFactory::getInstance().get(2);

    logger << "Please refer to the man pages for bincimap-up and bincimapd" 
	   << endl;
    logger << "for more information about how to invoke Binc IMAP." << endl;
    logger.flushContent();
  }
}

//------------------------------------------------------------------------
int main(int argc, char **argv) 
{
  IOFactory &iofactory = IOFactory::getInstance();
  iofactory.assign(1, new IO(stdout));
  iofactory.assign(2, new IO(stderr));

  IO &com = iofactory.get(1);
  IO &logger = iofactory.get(2);

  Session &session = Session::getInstance();

  session.importFromEnv();

  // log settings
  session.globalconfig.setSection("Log");
  string ipenv = session.globalconfig["ip environment variable"];
  // Initialize logger
  string ip = getenv(ipenv.c_str()) ? getenv(ipenv.c_str()) :
    getenv("TCPREMOTEIP") ? getenv("TCPREMOTEIP") :
    getenv("REMOTEIP") ? getenv("REMOTEIP") :
    getenv("SSLREMOTEIP") ? getenv("SSLREMOTEIP") : "?";
  session.setIP(ip);

  if (session.globalconfig["type"] == "multilog") {
    logger.setLogPrefix(session.getUserID() + "@" + ip + ":");
    logger.enableLogPrefix();
  } else if (session.globalconfig["type"] == "stderr" ) {
    // stderr is the default
  } else if (session.globalconfig["type"] == "" 
	     || session.globalconfig["type"] == "syslog") {
    session.globalconfig.setSection("Log");
    const string f = session.globalconfig["syslog facility"];
    const string fn = session.globalconfig["syslog facility number"];

    int facility;

    if (fn != "") facility = atoi(fn);
    else {
      if (f == "LOG_USER") facility = LOG_USER;
      else if (f == "LOG_LOCAL0") facility = LOG_LOCAL0;
      else if (f == "LOG_LOCAL1") facility = LOG_LOCAL1;
      else if (f == "LOG_LOCAL2") facility = LOG_LOCAL2;
      else if (f == "LOG_LOCAL3") facility = LOG_LOCAL3;
      else if (f == "LOG_LOCAL4") facility = LOG_LOCAL4;
      else if (f == "LOG_LOCAL5") facility = LOG_LOCAL5;
      else if (f == "LOG_LOCAL6") facility = LOG_LOCAL6;
      else if (f == "LOG_LOCAL7") facility = LOG_LOCAL7;
      else facility = LOG_DAEMON;
    }

    session.globalconfig.insert("syslog facility number", toString(facility));

    logger.setModeSyslog("bincimapd", facility);
  }

  // Now that we know the log type, we can flush.
  logger.flushContent();
  logger.flushOnEndl();

  // Read command line arguments
  try {
    session.parseCommandLine(argc, argv);
  } catch (exception &e) {
    logger << "Error reading command line arguments, "
	   << endl;
    logger.flushContent();
    return 111;
  }

  // Show help if asked for it
  if (session.command.help) {
    usage(argv[0]);
    logger.flushContent();
    return 111;
  }

  // Show help if asked for it
  if (session.command.version) {
    logger << "Binc IMAP v" << VERSION << endl;
    logger.flushContent();
    return 111;
  }

  char *logindetails = getenv("BINCIMAP_LOGIN");
  if (logindetails == NULL) {
    logger 
      << "BINCIMAP_LOGIN missing from environment (are you sure you invoked " 
      << argv[0] << " properly?)" << endl;
    logger.flushContent();
    return 111;
  }

  //-----

  // try to read local settings. if they are not set, then create a
  // local settings file.
  try {
    session.localconfig.load(".bincimap");
  } catch (StorageException &e) {
    // ignore it
  }

  session.globalconfig.setSection("Mailbox");
  session.localconfig.setSection("Mailbox");

  string tmp;
  if ((tmp = session.localconfig["depot"]) != "")
    session.globalconfig.insert("depot", tmp);
  if ((tmp = session.localconfig["path"]) != "")
    session.globalconfig.insert("path", tmp);
  if ((tmp = session.localconfig["type"]) != "")
    session.globalconfig.insert("type", tmp);

  DepotFactory &depotfactory = DepotFactory::getInstance();
  depotfactory.assign(new IMAPdirDepot());
  depotfactory.assign(new MaildirPPDepot());

  string depottype = session.globalconfig["depot"];
  if (depottype == "") depottype = "Maildir++";

  Depot *dep = depotfactory.get(depottype);
  if (dep == NULL) {
    logger << "Found no Depot for " << session.globalconfig["depot"]
	   << endl;
    return 111;
  }

  Depot &depot = *dep;
  depot.assign(new Maildir());
  depot.setDefaultType("Maildir");

  BrokerFactory brokerfactory = BrokerFactory::getInstance();

  brokerfactory.assign(new AppendOperator());
  brokerfactory.assign(new CapabilityOperator());
  brokerfactory.assign(new CheckOperator());
  brokerfactory.assign(new CloseOperator());
  brokerfactory.assign(new CopyOperator());
  brokerfactory.assign(new CreateOperator());
  brokerfactory.assign(new DeleteOperator());
  brokerfactory.assign(new ExamineOperator());
  brokerfactory.assign(new ExpungeOperator());
  brokerfactory.assign(new FetchOperator());
  brokerfactory.assign(new ListOperator());
  brokerfactory.assign(new LogoutOperator());
  brokerfactory.assign(new LsubOperator());
  brokerfactory.assign(new NoopPendingOperator());
  brokerfactory.assign(new RenameOperator());
  brokerfactory.assign(new SearchOperator());
  brokerfactory.assign(new SelectOperator());
  brokerfactory.assign(new StatusOperator());
  brokerfactory.assign(new StoreOperator());
  brokerfactory.assign(new SubscribeOperator());
  brokerfactory.assign(new UnsubscribeOperator());

  string path = session.globalconfig["path"];
  if (path == "") path = ".";
  else if (chdir(path.c_str()) != 0) {
    mkdir(path.c_str(), 0777);
    if (chdir(path.c_str()) != 0) {
      logger << "when entering depot " + toImapString(path) + ": "
	     << strerror(errno) << endl;
      return 111;
    }
  }

  if (depot.get("INBOX") == NULL) {
    if (session.globalconfig["auto create inbox"] == "yes")
      try {
	depot.createMailbox("INBOX");
      } catch (StorageException &e) {
	logger << "when creating INBOX: " << e.getReason() << endl;
	return 111;
      }
  }

  // load subscription list
  try {
    Storage subs;
    subs.load("bincimap-subscribed");
    subs.setSection("subscribed");

    string folder;
    string key;
    if (subs.getFirstEntry(key, folder))
      do {
	session.subscribed.push_back(folder);
      } while (subs.getNextEntry(key, folder));

  } catch (StorageException &e) {
    session.globalconfig.setSection("Mailbox");
    string autosubmailboxes = session.globalconfig["auto subscribe mailboxes"];
    vector<string> mboxes;
    split(autosubmailboxes, ",", mboxes);

    Storage subs;
    subs.setSection("subscribed");
    int n = 0;
    for (vector<string>::const_iterator i = mboxes.begin(); i != mboxes.end();
	 ++i) {
      string tmp = *i;
      trim(tmp);
      subs.insert(toString(n), tmp);
      session.subscribed.push_back(tmp);
      ++n;
    }

    try {
      subs.save("bincimap-subscribed");
    } catch (StorageException &e) {
      logger << "Saving bincimap-subscribed failed: " << e.getReason() << endl;
    }
  }

  session.setState(Session::AUTHENTICATED);

  session.globalconfig.setSection("Modules");

  // Loadable modules 

  // Not implemented yet -- use absolute paths for modules
  // vector<string> tmp_mods;
  // split(session.globalconfig["module path"], ":", tmp_mods);

  string k, v;
  if (session.globalconfig.getFirstEntry(k, v)) {
    do {
      try {
	if (k.substr(0,14) == "define module ") {
	  BincModule module(v);
	  modules.push_back(module);
	  module.load(brokerfactory);
	}
      }
      catch (const ModuleException &e) {
	logger << e.what() << endl;
      }
    } while (session.globalconfig.getNextEntry(k, v));
  }

  // Read timeout settings from global config
  session.globalconfig.setSection("Session");
  int idletimeout = atoi(session.globalconfig["idle timeout"]);
  int authtimeout = atoi(session.globalconfig["auth timeout"]);
  int transfertimeout = atoi(session.globalconfig["transfer timeout"]);

  // Settings are not allowed to break the IMAP protocol
  if (idletimeout < (30 * 60)) idletimeout = 30 * 60;

  // No auth timeout is not allowed.
  if (authtimeout < 30) authtimeout = 30;

  // Set transfer timeout
  com.setTransferTimeout(transfertimeout);
  logger.setTransferTimeout(transfertimeout);

  // Read transfer buffer size
  int buffersize = atoi(session.globalconfig["transfer buffer size"]);
  com.setBufferSize(buffersize >= 0 ? buffersize : 0);

  // umask settings
  session.globalconfig.setSection("Mailbox");
  unsigned int mode;
  sscanf(session.globalconfig["umask"].c_str(), "%o", &mode);
  umask((mode_t) mode);

  // Initialize timeout handler
  signal(SIGALRM, alarmhandler);

  const string details = logindetails;
  string::size_type det = details.find('+');
  if (det == string::npos) {
    logger << "invalid content of BINCIMAP_LOGIN - did you invoke "
       << argv[0] << " correctly?" << endl;
    return 111;
  }

  try {
    const string tag = details.substr(det + 1);
    const string command = details.substr(0, det);
    com << tag << " OK " << command << " completed" << endl;
    com.flushContent();
  } catch (InputException &e) {
    logger << "input error: " << e.getReason() << endl;
    return 111;
  }

  logger << "User <" << session.getUserID() << "> entered authenticated mode." << endl;

  logger.flushContent();

  // Read commands and run functions
  while (session.getState() != Session::LOGOUT) {
    Command command;
    bool recovery = false;
    try {
      Broker &broker = brokerfactory.getBroker(session.getState());
      
      com.noFlushOnEndl();

      alarm(session.getState() == Session::NONAUTHENTICATED 
	    ? authtimeout
	    : idletimeout);

      com.resetInput();
      com.enableInputLimit();
      broker.parse(command);
      alarm(0);

      logger.flushContent();
      
      session.addStatement();

      broker.process(depot, command);

      com << command.getTag() << " OK " << command.getName() 
	  << " completed" << endl;

    } catch (CriticalException &e) {
      logger << "Critical exception: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (OperatorException &e) {
      if (e.getCode() != Operator::OK && e.getCode() != Operator::NOTHING) {
	logger << "Error when processing " << command.getName() << ": " << e.getReason() << endl;
	logger.flushContent();
      }

      if (e.getCode() != Operator::NOTHING) {
	com << command.getTag();

	if (e.getCode() == Operator::OK) com << " OK ";
	else if (e.getCode() == Operator::BAD) { com << " BAD "; }
	else if (e.getCode() == Operator::NO) com << " NO ";
	com << e.getReason() << endl;
      }

    } catch (StorageException &e) {
      logger << "Storage warning, shutting down: " << e.getReason() << endl;
      com << "* BYE Binc IMAP shutting down: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (SyntaxException &e) {
      logger << "Error from client: " << e.getReason() << endl;
      com << "* NO Syntax Error: " << e.getReason() << endl;
      recovery = true;

    } catch (InputException &e) {
      logger << "Input error, shutting down: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (BincException &e) {
      logger << "Error: " << e.getReason() << endl;
      session.setState(Session::LOGOUT);

    } catch (...) {
      logger << "An unknown error occurred." << endl;
      session.setState(Session::LOGOUT);
    }

    com.flushContent();

    /* recovery after syntax error */    
    if (recovery)
      try {
	while (com.readChar() != '\n');
      } catch (InputException &e) {
	logger << "io warning, " << e.getReason() << endl;
	session.setState(Session::LOGOUT);
      }
  }

  logger << "Shutting down - bodies:" << session.getBodies()
	 << " statements:" << session.getStatements() << endl;

  com.flushContent();

  return 0;
}
