/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/depot.cc
 *  
 *  Description:
 *    <--->
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#include <map>
#include <string>

#include "depot.h"
#include "exceptions.h"
#include "mailbox/mailbox.h"
#include "mailbox/status.h"

using namespace ::std;
using namespace Binc;

//--------------------------------------------------------------------
DepotFactory::DepotFactory(void)
{
}

//--------------------------------------------------------------------
DepotFactory::~DepotFactory(void)
{
  //  for (vector<Depot *>::iterator i = depots.begin(); i != depots.end();
  //       ++i)
  //    delete *i;
}

//--------------------------------------------------------------------
Depot *DepotFactory::get(const string &name) const
{
  for (vector<Depot *>::const_iterator i = depots.begin(); i != depots.end();
       ++i)
    if ((*i)->getName() == name)
      return *i;

  return NULL;
}

//--------------------------------------------------------------------
void DepotFactory::assign(Depot *depot)
{
  depots.push_back(depot);
}

//--------------------------------------------------------------------
DepotFactory &DepotFactory::getInstance(void)
{
  static DepotFactory depotfactory;
  return depotfactory;
}

//--------------------------------------------------------------------
Depot::Depot(void) : enditerator(NULL, NULL)
{
  defaultmailbox = NULL;
  selectedmailbox = NULL;

  delimiter = '/';
}

//--------------------------------------------------------------------
Depot::Depot(const string &name) : enditerator(NULL, NULL)
{
  defaultmailbox = NULL;
  selectedmailbox = NULL;

  delimiter = '/';

  this->name = name;
}

//--------------------------------------------------------------------
Depot::~Depot(void)
{
}

//--------------------------------------------------------------------
void Depot::assign(Mailbox *m)
{
  for (vector<Mailbox *>::const_iterator i = backends.begin();
       i != backends.end(); ++i)
    if (*i == m) break;

  backends.push_back(m);
}

//--------------------------------------------------------------------
Mailbox *Depot::get(const string &s_in) const
{
  for (vector<Mailbox *>::const_iterator i = backends.begin();
       i != backends.end(); ++i)
    if ((*i)->isMailbox(mailboxToFilename(s_in)))
      return *i;

  return NULL;
}

//--------------------------------------------------------------------
void Depot::setSelected(Mailbox *m) throw (CriticalException &)
{
  for (vector<Mailbox *>::const_iterator i = backends.begin();
       i != backends.end(); ++i)
    if (*i == m) {
      selectedmailbox = m;
      return;
    }

  throw CriticalException("attempt to select unregistered Mailbox type");
}

//--------------------------------------------------------------------
const string &Depot::getName(void) const
{
  return name;
}

//--------------------------------------------------------------------
void Depot::setDelimiter(char c)
{
  delimiter = c;
}

//--------------------------------------------------------------------
const char Depot::getDelimiter(void) const
{
  return delimiter;
}

//--------------------------------------------------------------------
void Depot::setDefaultType(const string &name) throw (CriticalException &)
{
  for (vector<Mailbox *>::const_iterator i = backends.begin();
       i != backends.end(); ++i)
    if ((*i)->getName() == name) {
      defaultmailbox = *i;
      return;
    }

  throw CriticalException("attempt to default to unregistered Mailbox type "
			  + name);
}

//--------------------------------------------------------------------
Mailbox *Depot::getSelected(void) const
{
  return selectedmailbox;
}

//--------------------------------------------------------------------
Mailbox *Depot::getDefault(void) const
{
  return defaultmailbox;
}

//--------------------------------------------------------------------
int Depot::createMailbox(const string &s_in) const
  throw (CriticalException &, StorageException &)
{
  const string &mailboxname = mailboxToFilename(toCanonMailbox(s_in));
  if (mailboxname == "")
    throw StorageException("failed to create " + toImapString(s_in));

  Mailbox *mailbox = getDefault();
  if (mailbox == NULL)
    throw CriticalException("no default mailbox defined");

  mailbox->imapCreate(mailboxname, 0777);
  return 0;
}

//--------------------------------------------------------------------
int Depot::deleteMailbox(const string &s_in) const
  throw (CriticalException &, StorageException &)
{
  const string &mailboxname = mailboxToFilename(toCanonMailbox(s_in));

  Mailbox *mailbox = get(s_in);
  if (mailbox == NULL)
    throw StorageException(s_in + ": no such mailbox");
  
  mailbox->imapDelete(mailboxname);
  return 0;
}

//--------------------------------------------------------------------
int Depot::renameMailbox(const string &s_in, const string &t_in) const
  throw (StorageException &)
{
  const string &source = mailboxToFilename(s_in).c_str();
  const string &dest = mailboxToFilename(t_in).c_str();

  int nrenamed = 0;
  const iterator e = end();
  for (iterator i = begin("."); i != e; ++i) {
    string entry = *i;

    if (entry.substr(0, source.length()) == source) {
      string sourcename, destname;

      if (entry.length() == source.length()) {
	sourcename = source;
	destname = dest;

      } else if (entry.length() > source.length() 
		 && entry[source.length()] == '.') {
	sourcename = entry;
	destname = dest + entry.substr(dest.length());
      } else continue;

      if (rename(sourcename.c_str(), destname.c_str()) != 0) {
      } else
	nrenamed++;
      
      Mailbox *mailbox;
      if ((mailbox = get(filenameToMailbox(sourcename))) != NULL)
	mailbox->bumpUidValidity(filenameToMailbox(sourcename));
      if ((mailbox = get(filenameToMailbox(destname))) != NULL)
	mailbox->bumpUidValidity(filenameToMailbox(destname));
    }
  }

  if (nrenamed == 0) {
    IO &logger = IOFactory::getInstance().get(2);

    logger << "RENAME INFO: failed when renaming " << s_in
	   << " to " << t_in << "." << endl;
    throw StorageException("error when renaming " 
			   + toImapString(s_in)
			   + " to " + toImapString(t_in));
  }

  return 0;
}

//--------------------------------------------------------------------
Status &Depot::getStatus(const std::string &s_in) const
{
  const string mailbox = toCanonMailbox(s_in);
  Mailbox *m = get(mailbox);
  if (m == NULL) 
    throw StorageException("invalid mailbox " + toImapString(s_in));

  int statusid = m->getStatusID(mailboxToFilename(mailbox));
  if (mailboxstatuses.find(mailbox) != mailboxstatuses.end()) {
    Status &s = mailboxstatuses[mailbox];

    if (s.getStatusID() == statusid)
      return s;
  }

  Status s;
  if (!m->getStatus(mailboxToFilename(mailbox), s))
    throw StorageException("unable to get status of " + toImapString(s_in));

  s.setStatusID(statusid);
  
  return mailboxstatuses[mailbox] = s;
}

//--------------------------------------------------------------------
Depot::iterator::iterator(void)
{
  dirp = NULL;
}

//--------------------------------------------------------------------
Depot::iterator::iterator(DIR *dp, struct dirent *sp)
{
  dirp = dp;
  direntp = sp;
}

//--------------------------------------------------------------------
string Depot::iterator::operator * (void) const
{
  if (direntp == NULL)
    throw CriticalException("* operator called on NULL direntp");

  return direntp->d_name;
}

//--------------------------------------------------------------------
void Depot::iterator::operator ++ (void)
{
  direntp = readdir(dirp);
}

//--------------------------------------------------------------------
bool Depot::iterator::operator == (Depot::iterator i) const
{
  return direntp == i.direntp;
}

//--------------------------------------------------------------------
bool Depot::iterator::operator != (Depot::iterator i) const
{
  return direntp != i.direntp;
}

//--------------------------------------------------------------------
Depot::iterator Depot::begin(const string &path) const
{
  Depot::iterator i;

  if ((i.dirp = opendir(path.c_str())) == NULL)
    throw BincException("opendir on " + path + " failed");

  ++i;
  return i;
}

//--------------------------------------------------------------------
const Depot::iterator &Depot::end(void) const
{
  return enditerator;
}

//--------------------------------------------------------------------
MaildirPPDepot::MaildirPPDepot(void) : Depot("Maildir++")
{
}

//--------------------------------------------------------------------
MaildirPPDepot::~MaildirPPDepot(void)
{
}

//--------------------------------------------------------------------
string MaildirPPDepot::mailboxToFilename(const string &m) const
{
  string prefix = "INBOX"; prefix += delimiter;

  string mm = m;
  trim(mm, string(&delimiter, 1));

  if (mm == "INBOX") return ".";
  else if (mm.length() <= 6) return "";
  else if (mm.substr(0, 6) != prefix) return "";
  else if (mm.find("..") != string::npos) return "";
  else if (mm != "" && mm.substr(1).find(".") != string::npos) return "";
  else {
    string tmp = mm.substr(6);
    for (string::iterator i = tmp.begin(); i != tmp.end(); ++i)
      if (*i == '/') *i = '.';

    return "." + tmp;
  }
}

//--------------------------------------------------------------------
string MaildirPPDepot::filenameToMailbox(const string &m) const
{
  if (m == ".") return "INBOX";
  else if (m.find(delimiter) != string::npos) return "";
  else if (m != "" && m[0] == '.') {
    string tmp = m;
    for (string::iterator i = tmp.begin(); i != tmp.end(); ++i)
      if (*i == '.') *i = delimiter;

    return "INBOX" + tmp;
  } else return "";
}

//--------------------------------------------------------------------
IMAPdirDepot::IMAPdirDepot(void) : Depot("IMAPdir")
{
}

//--------------------------------------------------------------------
IMAPdirDepot::~IMAPdirDepot(void)
{
}

//--------------------------------------------------------------------
string IMAPdirDepot::mailboxToFilename(const string &m) const
{
  string tmp;
  string mm = m;
  trim(mm, string(&delimiter, 1));

  string::const_iterator i = mm.begin();
  while (i != mm.end()) {
    if (*i == delimiter) {
      tmp += '.';
    } else if (*i == '\\') {
      tmp += "\\\\";
    } else if (*i == '.') {
      if (i == mm.begin())
	tmp += ".";
      else
	tmp += "\\.";
    } else {
      tmp += *i;
    }

    ++i;
  }

  return tmp;
}

//--------------------------------------------------------------------
string IMAPdirDepot::filenameToMailbox(const string &m) const
{
  string tmp;
  bool escape = false;

  string::const_iterator i = m.begin();
  while (i != m.end()) {
    if (*i == '.') {
      if (i != m.begin() && !escape) tmp += delimiter;
      else if (i == m.begin() || escape) tmp += '.';
      escape = false;
    } else if (*i == '\\') {
      if (!escape) escape = true; else {
	tmp += '\\';
	escape = false;
      }
    } else if (*i == delimiter) {
      return "";
    } else {
      if (escape) return "";
      else {
	tmp += *i;
	escape = false;
      }
    }

    ++i;
  }

  return tmp;
}
