/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/io/io-ssl.cc
 *  
 *  Description:
 *    Implementation of the IO class.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef WITH_SSL

#include <string>

#include <openssl/ssl.h>

#include "../session.h"
#include "io-ssl.h"

using namespace ::std;
using namespace Binc;

//------------------------------------------------------------------------
SSLEnabledIO::SSLEnabledIO(void)
{
}

//------------------------------------------------------------------------
SSLEnabledIO::SSLEnabledIO(FILE *fp) : IO(fp)
{
}

//------------------------------------------------------------------------
SSLEnabledIO::~SSLEnabledIO(void)
{
  switch (mode) {
  case MODE_SSL:
    SSL_shutdown(ssl);
    SSL_shutdown(ssl);
    break;
  case MODE_SYSLOG:
    closelog();
    break;
  default:
    break;
  }
}


//------------------------------------------------------------------------
void SSLEnabledIO::setModeSSL(void)
{
  SSL_library_init();
  SSL_load_error_strings();

  OpenSSL_add_ssl_algorithms();
  
  SSL_CTX *ctx = SSL_CTX_new(SSLv23_server_method());
  if (ctx == NULL)
    throw InputException("SSL error: internal error when creating CTX: " 
			 + string(ERR_error_string(ERR_get_error(), NULL)));
  
  SSL_CTX_set_options(ctx, SSL_OP_ALL);
  SSL_CTX_sess_set_cache_size(ctx, 1);
  SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_SERVER
				 | SSL_SESS_CACHE_NO_AUTO_CLEAR
				 | SSL_SESS_CACHE_NO_INTERNAL_LOOKUP);

  Session &session = Session::getInstance();

  session.globalconfig.setSection("SSL");
  string clist = session.globalconfig["cipher list"];;
  if (clist == "")
    throw InputException("SSL error: cipher list undefined");

  if (!SSL_CTX_set_cipher_list(ctx, clist.c_str()))
    throw InputException("SSL error: cannot load cipher list " + clist);

  session.globalconfig.setSection("SSL");
  string CAfile = session.globalconfig["ca file"];
  if (CAfile == "") CAfile == "/usr/share/ssl/certs/.crt";

  SSL_CTX_set_default_verify_paths(ctx);

  string pemname = session.globalconfig["pem file"];
  if (pemname == "") pemname = "/usr/share/ssl/certs/stunnel.pem";

  if (!SSL_CTX_use_certificate_file(ctx, pemname.c_str(), SSL_FILETYPE_PEM))
    throw InputException ("SSL error: unable to use certificate in PEM file: "
			  + pemname + ": "
			  + string(ERR_error_string(ERR_get_error(), NULL)));

  if (!SSL_CTX_use_PrivateKey_file(ctx, pemname.c_str(), SSL_FILETYPE_PEM))
    throw InputException ("SSL error: unable to use private key in PEM file: "
			  + pemname + ": "
			  + string(ERR_error_string(ERR_get_error(), NULL)));

  if (!SSL_CTX_check_private_key(ctx))
    throw InputException ("SSL error: can't check private key in PEM file: " 
			  + pemname + ": "
			  + string(ERR_error_string(ERR_get_error(), NULL)));

  if (session.globalconfig["verify peer"] == "yes")
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE, NULL);
  else
    SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);

  SSL_CTX_set_client_CA_list(ctx, SSL_load_client_CA_file(CAfile.c_str()));

  SSL *ssl = SSL_new(ctx);
  if (ssl == NULL)
    throw InputException ("SSL error: when creating SSL object: "
			  + string(ERR_error_string(ERR_get_error(), NULL)));

  SSL_clear(ssl);

  SSL_set_rfd(ssl, 0);
  SSL_set_wfd(ssl, 1);
  SSL_set_accept_state(ssl);
   
  fflush(stdout);

  int result = SSL_accept(ssl);
  if (result == 1) {
    SSL_CIPHER* cipher;
    const char *tls_protocol = NULL;
    const char *tls_cipher_name = NULL;
    int tls_cipher_usebits = 0;
    int tls_cipher_algbits = 0;

    tls_protocol = SSL_get_version(ssl);
    cipher = SSL_get_current_cipher(ssl);
    tls_cipher_name = SSL_CIPHER_get_name(cipher);
    tls_cipher_usebits = SSL_CIPHER_get_bits(cipher, &tls_cipher_algbits);
    
  } else if (result == 0) {
    SSL_get_error(ssl, 0);
    throw InputException("SSL error: authentication failed: "
			 + string(ERR_error_string(ERR_get_error(), NULL)));
  } else if (result < 0)
    throw InputException("SSL error: " 
			 + string(ERR_error_string(ERR_get_error(), NULL)));

  this->ssl = ssl;
  mode = MODE_SSL;
}

//------------------------------------------------------------------------
bool SSLEnabledIO::isModeSSL(void)
{
  return mode == MODE_SSL;
}

//------------------------------------------------------------------------
void SSLEnabledIO::writeStr(const string s)
{
  if (s == "")
    return;
  
  if (mode == MODE_PLAIN) {
    // Set transfer timeout
    alarm(transfertimeout);
    int n = fwrite(s.c_str(), 1, s.length(), fpout);
    fflush(fpout);
    alarm(0);
    
    if (n != (int) s.length())
      throw InputException("fwrite wrote too few bytes");
  } else if (mode == MODE_SYSLOG) {
    // Set transfer timeout
    alarm(transfertimeout);
    syslog(LOG_INFO, "%s", s.c_str());
    alarm(0);
  } else if (mode == MODE_SSL) {
    do {
      alarm(transfertimeout);
      int retval = SSL_write(ssl, s.c_str(), 
			     s.length());
      alarm(0);
      
      if (retval > 0)
	break;
      else if (retval == 0) {
	/* call get_error */
	throw InputException("SSL_write returned 0");
      } else if (retval < 0) {
	int err = SSL_get_error(ssl, retval);
	
	if (err != SSL_ERROR_WANT_READ && err != SSL_ERROR_WANT_WRITE)
	  throw InputException("SSL_write returned < 0");
	else {
	  // retry when we get this error.
	  //if (err == SSL_ERROR_WANT_READ) {
	  // cerr << "SSL_ERROR_WANT_READ" << endl;
	  //} else {
	  // cerr << "SSL_ERROR_WANT_WRITE" << endl;
	  //}
	} 
      }
    } while (1);
  }
}

//------------------------------------------------------------------------
int SSLEnabledIO::readChar(int timeout, bool retry)
{
  int nextchar = 0;
  char c;
  
  if (inputBuffer.size() != 0) {
    c = inputBuffer.top();
    inputBuffer.pop();
    return c;
  }

  if (inputlimit && inputsize >= 8192)
    throw InputException("Input limit is 8192 characters.");
  
  if (mode == MODE_PLAIN) {
    fd_set rfds;
    FD_ZERO(&rfds);
    FD_SET(0, &rfds);

    int r = select(1, &rfds, NULL, NULL, timeout);
    if (r > 0) {
      if ((nextchar = read(fileno(stdin), &c, 1)) == 1) {
	nextchar = c;
	c = nextchar;
      } else
	throw InputException("readChar EOF");
    }
    else if (r == 0)
      throw TimeoutException("Reading from stdin timed out.");
    else
      throw InputException("Error reading from stdin");
  } else if (mode == MODE_SSL) {
    do {
      fd_set rfds;
      FD_ZERO(&rfds);
      FD_SET(0, &rfds);

      int r = select(1, &rfds, NULL, NULL, timeout);

      if (r > 0) {
	int retval = SSL_read(ssl, &c, 1);      
	if (retval > 0)
	  break;
	else if (retval == 0) {
	  /* call get_error */
	  throw InputException("SSL error: client disconnected");
	} else if (retval < 0) {
	  int err = SSL_get_error(ssl, retval);
	  if (err != SSL_ERROR_WANT_READ && err != SSL_ERROR_WANT_WRITE)
	    throw InputException("SSL error: read returned " 
				 + toString(retval)
				 + ": " 
				 + string(ERR_error_string(ERR_get_error(),
							   NULL)));
	  if (timeout == 0) 
	    throw TimeoutException("Reading from SSL timed out.");

	  if (!retry)
	    return -2;
	}
      }
      else if (r == 0)
	throw TimeoutException("Reading from SSL timed out.");
      else
	throw InputException("Error reading from SSL socket.");
    } while (1);
  }

  ++inputsize;
  
  if (nextchar == EOF)
    throw InputException("readChar EOF");
  else
    return c;
}

//------------------------------------------------------------------------
int SSLEnabledIO::pending(void) const
{
  return (mode == MODE_PLAIN) ?  0 : SSL_pending(ssl);
}

#endif
