/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-search.cc
 *  
 *  Description:
 *    Implementation of the SEARCH command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <iostream>
#include <algorithm>

#include <ctype.h>
    
#include "parsers/imap/imapparser.h"
#include "mailbox/mailbox.h"
#include "parsers/mime/mime.h"
#include "io/io.h"
#include "util/convert.h"

#include "parsers/imap/recursivedescent/recursivedescent.h"

#include "session.h"
#include "depot.h"
#include "operators.h"

using namespace ::std;
using namespace Binc;

namespace {

  static stack<char> CRLFbuffer;
  static char CRLFlastchar;

  // read one character from the input file. pass the data as a pure
  // CRLF text file, regardless of what the source is.
  inline int fgetcCRLF(FILE *fp)
  {
    int c;

    // if our stack has any characters in it, then return those until
    // the stack is empty.
    if (CRLFbuffer.size() != 0) {
      c = CRLFbuffer.top();
      CRLFbuffer.pop();
      CRLFlastchar = c;
      return c;
    }

    // read one character from input, store this in c. note that EOF
    // is also returned if found.
    while ((c = fgetc(fp)) != '\n') {
	// if the character read from the file was not an LF, then we
	// print it if it is not a CR.
	if (c != '\r') {
	  CRLFlastchar = c;
	  return c;
	}

	// if the character in the file was CR, then we skip to the
	// next character.
    }

    if (CRLFlastchar != '\r') {
      // if the previous character returned was not a CR, then push
      // the LF character on the stack and return the CR character
      // instead. obviously this messes up files with CRCRCRLF ->
      // this is converted to CRLF. LFLFLF, however, is
      // CRLFCRLFCRLF.
      CRLFbuffer.push('\n');
      CRLFlastchar = '\r';
      return '\r';
    }

    // if the previous character returned was a CR, then return
    // the LF.
    CRLFlastchar = '\n';
    return '\n';
  }


  inline bool compareStringToQueue(const std::string &s_in, char *bqueue,
				   int pos, int size)
  {
    if (s_in[0] != bqueue[pos % size]) return false;
    
    for (int i = 0; i < size; ++i)
      if (s_in.at(i) != bqueue[(pos + i) % size])
	return false;
    
    return true;
  }

}

bool SearchOperator::SearchNode::convertDate(const string &date, 
					     time_t &t,
					     const string &delim)
{
  vector<string> parts;
  split(date, delim, parts);
  if (parts.size() < 3) return false;
  
  struct tm mold;
  memset((char *) &mold, 0, sizeof(struct tm));
  mold.tm_mday = atoi(parts[0].c_str());
  mold.tm_year = atoi(parts[2].c_str()) - 1900;
  
  // accept mixed case months. this is more than the standard
  // accepts.
  string month = parts[1];
  lowercase(month);
  
  if (month == "jan") mold.tm_mon = 0;
  else if (month == "feb") mold.tm_mon = 1;
  else if (month == "mar") mold.tm_mon = 2;
  else if (month == "apr") mold.tm_mon = 3;
  else if (month == "may") mold.tm_mon = 4;
  else if (month == "jun") mold.tm_mon = 5;
  else if (month == "jul") mold.tm_mon = 6;
  else if (month == "aug") mold.tm_mon = 7;
  else if (month == "sep") mold.tm_mon = 8;
  else if (month == "oct") mold.tm_mon = 9;
  else if (month == "nov") mold.tm_mon = 10;
  else if (month == "dec") mold.tm_mon = 11;
  
  t = mktime(&mold);
  return true;
}

bool SearchOperator::SearchNode::convertDateHeader(const string &d_in, time_t &t)
{
  string date = d_in;
  string::size_type n = date.find(',');
  if (n != string::npos)
    date = date.substr(n + 1);
  trim(date);
  
  bool result = convertDate(date, t, " ");
  return result;
}


//----------------------------------------------------------------------
SearchOperator::SearchNode::SearchNode(void)
{
}

//----------------------------------------------------------------------
SearchOperator::SearchNode::SearchNode(const BincImapParserSearchKey &a)
{
  init(a);
}

//----------------------------------------------------------------------
int SearchOperator::SearchNode::getType(void) const
{
  return type;
}

//----------------------------------------------------------------------
bool SearchOperator::SearchNode::match(Mailbox *mailbox, Message *m, int seqnr, int lastmessage, int lastuid) const
{
  HeaderItem hitem;
  string tmp;

  switch (type) {
    //--------------------------------------------------------------------
  case S_ALL: 
    return true;
    //--------------------------------------------------------------------
  case S_ANSWERED: 
    return (m->getStdFlags() & Message::F_ANSWERED);
    //--------------------------------------------------------------------
  case S_BCC: {
    mailbox->parseOnlyHeader(*m);
    if (m->h.getFirstHeader("bcc", hitem)) {
      string tmp = hitem.getValue();
      uppercase(tmp);
      string a = astring;
      uppercase(a);
      return (tmp.find(a) != string::npos);
    } else
      return false;
  } //--------------------------------------------------------------------
  case S_BEFORE: {
    time_t mtime = (time_t) atoi(m->getID().c_str());
    struct tm *mtime_ = localtime(&mtime);
    mtime_->tm_sec = 0;
    mtime_->tm_min = 0;
    mtime_->tm_hour = 0;
    mtime_->tm_wday = 0;
    mtime_->tm_yday = 0;
    mtime_->tm_isdst = 0;
    mtime = mktime(mtime_);

    time_t atime;
    if (!convertDate(date, atime)) {
      IO &logger = IOFactory::getInstance().get(2);

      logger << "warning, unable to convert " << date << 
	" to a time_t" << endl;
      return false;
    }

    return mtime < atime;
  } //--------------------------------------------------------------------
  case S_BODY: {
    mailbox->parseFull(*m);

    while (CRLFbuffer.size() != 0)
      CRLFbuffer.pop();
    CRLFlastchar = '\0';

    // search the body part of the message..
    FILE *fp;
    if ((fp = mailbox->fopen(m->getID())) != NULL) {

      fseek(fp, 0, SEEK_SET);

      int c = '\0';
      for (int i = 0; i < m->getBodyStartOffset(); ++i)
	if ((c = fgetcCRLF(fp)) == EOF) break;

      char *ring = new char[astring.length()];
      int pos = 0;
      int length = m->getBodyLength();
      while ((c = fgetcCRLF(fp)) != EOF && length--) {
	ring[pos % astring.length()] = toupper(c);

	if (compareStringToQueue(astring, ring, pos + 1, astring.length())) {
	  fclose(fp);
	  delete ring;
	  return true;
	}

	++pos;
      }

      delete ring;
      fclose(fp);
    } else {
      // FIXME: *** ERROR OPENING FILE ***
    }

    return false;

  } //--------------------------------------------------------------------
  case S_CC: {
    mailbox->parseOnlyHeader(*m);
    if (m->h.getFirstHeader("cc", hitem)) {
      string tmp = hitem.getValue();
      uppercase(tmp);
      string a = astring;
      uppercase(a);
      return (tmp.find(a) != string::npos);
    } else
      return false;
  } //--------------------------------------------------------------------
  case S_DELETED:
    return (m->getStdFlags() & Message::F_DELETED);
    //--------------------------------------------------------------------
  case S_FLAGGED:
    return (m->getStdFlags() & Message::F_FLAGGED);
    //--------------------------------------------------------------------
  case S_FROM: {
    mailbox->parseOnlyHeader(*m);
    if (m->h.getFirstHeader("from", hitem)) {
      string tmp = hitem.getValue();
      uppercase(tmp);
      string a = astring;
      uppercase(a);
      return (tmp.find(a) != string::npos);
    } else
      return false;
  } //--------------------------------------------------------------------
  case S_KEYWORD: 
    // the server does not support keywords
    return false;
    //--------------------------------------------------------------------
  case S_NEW:
    return (m->getStdFlags() & Message::F_RECENT) 
      && !(m->getStdFlags() & Message::F_SEEN);
    //--------------------------------------------------------------------
  case S_OLD:
    return !(m->getStdFlags() & Message::F_RECENT);
    //--------------------------------------------------------------------
  case S_ON: {
    time_t mtime = (time_t) atoi(m->getID().c_str());
    struct tm *mtime_ = localtime(&mtime);
    mtime_->tm_sec = 0;
    mtime_->tm_min = 0;
    mtime_->tm_hour = 0;
    mtime_->tm_wday = 0;
    mtime_->tm_yday = 0;
    mtime_->tm_isdst = 0;
    mtime = mktime(mtime_);

    time_t atime;
    if (!convertDate(date, atime)) {
      IO &logger = IOFactory::getInstance().get(2);

      logger << "warning, unable to convert " << date << 
	" to a time_t" << endl;
      return false;
    }

    return mtime == atime;
  } //--------------------------------------------------------------------
  case S_RECENT:
    return (m->getStdFlags() & Message::F_RECENT);
    //--------------------------------------------------------------------
  case S_SEEN:
    return (m->getStdFlags() & Message::F_SEEN);
    //--------------------------------------------------------------------
  case S_SINCE: {
    time_t mtime = (time_t) atoi(m->getID().c_str());
    struct tm *mtime_ = localtime(&mtime);
    mtime_->tm_sec = 0;
    mtime_->tm_min = 0;
    mtime_->tm_hour = 0;
    mtime_->tm_wday = 0;
    mtime_->tm_yday = 0;
    mtime_->tm_isdst = 0;
    mtime = mktime(mtime_);

    time_t atime;
    if (!convertDate(date, atime)) {
      IO &logger = IOFactory::getInstance().get(2);

      logger << "warning, unable to convert " << date << 
	" to a time_t" << endl;
      return false;
    }

    return mtime >= atime;
  } //--------------------------------------------------------------------
  case S_SUBJECT: {
    mailbox->parseOnlyHeader(*m);
    if (m->h.getFirstHeader("subject", hitem)) {
      string tmp = hitem.getValue();
      uppercase(tmp);
      string a = astring;
      uppercase(a);
      return (tmp.find(a) != string::npos);
    } else
      return false;
  } //--------------------------------------------------------------------
  case S_TEXT: {
    // search the body part of the message..
    FILE *fp;
    if ((fp = mailbox->fopen(m->getID())) != NULL) {
      // FIXME: *** ERROR OPENING FILE ***
      int c;
      char *ring = new char[astring.length()];
      int pos = 0;
      while ((c = fgetc(fp)) != EOF) {
	ring[pos % astring.length()] = toupper(c);

	if (compareStringToQueue(astring, ring, pos + 1, astring.length())) {
	  fclose(fp);
	  delete ring;
	  return true;
	}

	++pos;
      }

      delete ring;
      fclose(fp);
    }

    return false;
  } //--------------------------------------------------------------------
  case S_TO: {
    mailbox->parseOnlyHeader(*m);
    if (m->h.getFirstHeader("to", hitem)) {
      string tmp = hitem.getValue();
      uppercase(tmp);
      string a = astring;
      uppercase(a);
      return (tmp.find(a) != string::npos);
    } else
      return false;
  }  //--------------------------------------------------------------------
  case S_UNANSWERED:
    return !(m->getStdFlags() & Message::F_ANSWERED);
    //--------------------------------------------------------------------
  case S_UNDELETED:
    return !(m->getStdFlags() & Message::F_DELETED);
    //--------------------------------------------------------------------
  case S_UNFLAGGED:
    return !(m->getStdFlags() & Message::F_FLAGGED);
    //--------------------------------------------------------------------
  case S_UNKEYWORD:
    // the server does not support keywords
    return true;
    //--------------------------------------------------------------------
  case S_UNSEEN:
    return !(m->getStdFlags() & Message::F_SEEN);
    //--------------------------------------------------------------------
  case S_DRAFT:
    return (m->getStdFlags() & Message::F_DRAFT);
    //--------------------------------------------------------------------
  case S_HEADER: {
    mailbox->parseOnlyHeader(*m);

    string a = astring;
    lowercase(a);
    if (m->h.getFirstHeader(a, hitem)) {
      string tmp = hitem.getValue();
      uppercase(tmp);
      string b = bstring;
      uppercase(b);
      return (tmp.find(b) != string::npos);
    } else
      return false;
  } //--------------------------------------------------------------------
  case S_LARGER:
    if (m->getSize() == -1)
      mailbox->parseFull(*m);

    return (m->getSize() > number);
    //--------------------------------------------------------------------
  case S_NOT:
    for (vector<SearchNode>::const_iterator i = children.begin();
	 i != children.end(); ++i)
      if ((*i).match(mailbox, m, seqnr, lastmessage, lastuid))
	return false;
    return true;
    //--------------------------------------------------------------------
  case S_OR:
    for (vector<SearchNode>::const_iterator i = children.begin();
	 i != children.end(); ++i)
      if ((*i).match(mailbox, m, seqnr, lastmessage, lastuid))
	return true;
    return false;
    //--------------------------------------------------------------------
  case S_SENTBEFORE: {
    mailbox->parseOnlyHeader(*m);

    if (!m->h.getFirstHeader("date", hitem))
      return false;

    string tmp = hitem.getValue();
    lowercase(tmp);

    time_t mtime;
    if (!convertDateHeader(tmp, mtime))
      return false;

    if (mtime == (time_t) -1)
      return false;

    time_t atime;
    if (!convertDate(date, atime)) {
      IO &logger = IOFactory::getInstance().get(2);

      logger << "warning, unable to convert " << date << 
	" to a time_t" << endl;
      return false;
    }

    return mtime < atime;
  } //--------------------------------------------------------------------
  case S_SENTON: {
    mailbox->parseOnlyHeader(*m);

    if (!m->h.getFirstHeader("date", hitem))
      return false;

    string tmp = hitem.getValue();
    lowercase(tmp);

    time_t mtime;
    if (!convertDateHeader(tmp, mtime))
      return false;

    if (mtime == (time_t) -1)
      return false;

    time_t atime;
    if (!convertDate(date, atime)) {
      IO &logger = IOFactory::getInstance().get(2);

      logger << "warning, unable to convert " << date << 
	" to a time_t" << endl;
      return false;
    }

    return mtime == atime;
  } //--------------------------------------------------------------------
  case S_SENTSINCE: {
    mailbox->parseOnlyHeader(*m);

    if (!m->h.getFirstHeader("date", hitem))
      return false;

    string tmp = hitem.getValue();
    lowercase(tmp);

    time_t mtime;
    if (!convertDateHeader(tmp, mtime))
      return false;

    if (mtime == (time_t) -1)
      return false;

    time_t atime;
    if (!convertDate(date, atime)) {
      IO &logger = IOFactory::getInstance().get(2);

      logger << "warning, unable to convert " << date << 
	" to a time_t" << endl;
      return false;
    }

    return mtime >= atime;
  } //--------------------------------------------------------------------
  case S_SMALLER:
    if (m->getSize() == -1)
      mailbox->parseFull(*m);

    return (m->getSize() < number);
    //--------------------------------------------------------------------
  case S_UID:
    if (!bset->isInSet(m->getUID()))
      if (!(m->getUID() == lastuid && !bset->isLimited()))
	return false;
    return true;
    //--------------------------------------------------------------------
  case S_UNDRAFT:
    return !(m->getStdFlags() & Message::F_DRAFT);
    //--------------------------------------------------------------------
  case S_SET:
    if (!bset->isInSet(seqnr))
      if (!(seqnr == lastmessage && !bset->isLimited()))
	return false;
    return true;
    //--------------------------------------------------------------------
  case S_AND:
    for (vector<SearchNode>::const_iterator i = children.begin();
	 i != children.end(); ++i)
      if (!(*i).match(mailbox, m, seqnr, lastmessage, lastuid))
	return false;
    return true;
  }
  return false;
}

//----------------------------------------------------------------------
void SearchOperator::SearchNode::init(const BincImapParserSearchKey &a)
{
  astring = a.astring;
  bstring = a.bstring;
  date = a.date;
  number = a.number;
  uppercase(astring);
  uppercase(bstring);
  uppercase(date);

  if (a.name      == "ALL")            { type = S_ALL;        weight = 1; }
  else if (a.name == "ANSWERED")       { type = S_ANSWERED;   weight = 1; }
  else if (a.name == "BCC")            { type = S_BCC;        weight = 2; }
  else if (a.name == "BEFORE")         { type = S_BEFORE;     weight = 2; }
  else if (a.name == "BODY")           { type = S_BODY;       weight = 1; }
  else if (a.name == "CC")             { type = S_CC;         weight = 2; }
  else if (a.name == "DELETED")        { type = S_DELETED;    weight = 1; }
  else if (a.name == "FLAGGED")        { type = S_FLAGGED;    weight = 1; }
  else if (a.name == "FROM")           { type = S_FROM;       weight = 2; }
  else if (a.name == "KEYWORD")        { type = S_KEYWORD;    weight = 3; }
  else if (a.name == "NEW")            { type = S_NEW;        weight = 1; }
  else if (a.name == "OLD")            { type = S_OLD;        weight = 1; }
  else if (a.name == "ON")             { type = S_ON;         weight = 1; }
  else if (a.name == "RECENT")         { type = S_RECENT;     weight = 1; }
  else if (a.name == "SEEN")           { type = S_SEEN;       weight = 1; }
  else if (a.name == "SINCE")          { type = S_SINCE;      weight = 1; }
  else if (a.name == "SUBJECT")        { type = S_SUBJECT;    weight = 2; }
  else if (a.name == "TEXT")           { type = S_TEXT;       weight = 4; }
  else if (a.name == "TO")             { type = S_TO;         weight = 2; }
  else if (a.name == "UNANSWERED")     { type = S_UNANSWERED; weight = 1; }
  else if (a.name == "UNDELETED")      { type = S_UNDELETED;  weight = 1; }
  else if (a.name == "UNFLAGGED")      { type = S_UNFLAGGED;  weight = 1; }
  else if (a.name == "UNKEYWORD")      { type = S_UNKEYWORD;  weight = 1; }
  else if (a.name == "UNSEEN")         { type = S_UNSEEN;     weight = 1; }
  else if (a.name == "DRAFT")          { type = S_DRAFT;      weight = 1; }
  else if (a.name == "HEADER")         { type = S_HEADER;     weight = 3; }
  else if (a.name == "LARGER")         { type = S_LARGER;     weight = 4; }
  else if (a.name == "NOT")            {
    // *******                         NOT
    type = S_NOT;
    weight = 1;

    vector<BincImapParserSearchKey>::const_iterator i = a.children.begin();
    while (i != a.children.end()) {
      SearchNode b(*i);
      weight += b.getWeight();
      children.push_back(b);
      ++i;
    }

  } else if (a.name == "OR") {
    // *******                         OR
    type = S_OR;
    weight = 0;

    vector<BincImapParserSearchKey>::const_iterator i = a.children.begin();
    while (i != a.children.end()) {
      SearchNode b(*i);
      weight += b.getWeight();

      children.push_back(b);
      ++i;
    }

  } else if (a.name == "SENTBEFORE")   { type = S_SENTBEFORE; weight = 1; }
  else if (a.name == "SENTON")         { type = S_SENTON;     weight = 1; }
  else if (a.name == "SENTSINCE")      { type = S_SENTSINCE;  weight = 1; }
  else if (a.name == "SMALLER")        { type = S_SMALLER;    weight = 4; } 
  else if (a.name == "UID") {
    bset = &a.getSet();
    type = S_UID;
    weight = 1;
  } else if (a.name == "UNDRAFT")        { type = S_UNDRAFT;    weight = 1; }
  else if (a.type == BincImapParserSearchKey::KEY_SET) {
    bset = &a.getSet();
    type = S_SET;
    weight = 1;
  } else if (a.type == BincImapParserSearchKey::KEY_AND) {
    // *******                         AND
    type = S_AND;
    weight = 0;

    vector<BincImapParserSearchKey>::const_iterator i = a.children.begin();
    while (i != a.children.end()) {
      SearchNode b(*i);
      weight += b.getWeight();
      children.push_back(b);
      ++i;
    }
  }
}

//----------------------------------------------------------------------
int SearchOperator::SearchNode::getWeight(void) const
{
  return weight;
}

//----------------------------------------------------------------------
void SearchOperator::SearchNode::setWeight(int i)
{
  weight = i;
}

//----------------------------------------------------------------------
void SearchOperator::SearchNode::order(void)
{
  for (vector<SearchNode>::iterator i = children.begin();
       i != children.end(); ++i)
    (*i).order();
  ::stable_sort(children.begin(), children.end(), compareNodes);
}

//----------------------------------------------------------------------
SearchOperator::SearchOperator(void)
{
}

//----------------------------------------------------------------------
SearchOperator::~SearchOperator(void)
{
}

//----------------------------------------------------------------------
const string SearchOperator::getName(void) const
{
  return "SEARCH";
}

//----------------------------------------------------------------------
int SearchOperator::getState(void) const
{
  return Session::SELECTED;
}

//------------------------------------------------------------------------
void SearchOperator::process(Depot &depot, Command &command) const
{
  IO &com = IOFactory::getInstance().get(1);
  IO &logger = IOFactory::getInstance().get(2);

  Mailbox *mailbox = depot.getSelected();
  if (mailbox == NULL)
    throw CriticalException("search when no mailbox is selected");

  if (command.getCharSet() != "" && command.getCharSet() != "US-ASCII")
    throw OperatorException(Operator::NO, "[BADCHARSET (\"US-ASCII\")]");

  com << "* SEARCH";

  SearchNode s(command.searchkey);
  s.order();

  const int maxsqnr = mailbox->getMaxSqnr();
  const int maxuid = mailbox->getMaxUid();

  Mailbox::iterator e = mailbox->end();
  for (Mailbox::iterator i = mailbox->begin(); i != e; ++i) {
    Message & message = *i;

    try {
      if (s.match(mailbox, &message, i.getSqnr(), maxsqnr, maxuid)) {
	com << " " << (command.getUidMode() ?  message.getUID() : i.getSqnr());
	com.flushContent();
      }
    } catch (MessageException &e) {
      logger << "warning: " << e.getReason() << endl;
    }
    message.clear();
  }

  com << endl;
}

bool SearchOperator::parse(Command & c_in) const
{
  try {
    string command;
    if (expectThisString("UID SEARCH")) {
      c_in.setUidMode();
      command = "UID SEARCH";
    } else if (expectThisString("SEARCH")) {
      command = "SEARCH";
    } else return false;

    if (!expectSPACE())
      throw SyntaxException("expected SPACE after " + command);

    if (expectThisString("CHARSET")) {
      if (!expectSPACE())
	throw SyntaxException("expected SPACE after CHARSET");

      string charset;
      if (!expectAstring(charset))
	throw SyntaxException("expected astring after CHARSET SPACE");
      c_in.setCharSet(charset);

      if (!expectSPACE())
	throw SyntaxException("expected SPACE after CHARSET SPACE astring");
    }

    BincImapParserSearchKey b;
    if (!expectSearchKey(b))
      throw SyntaxException("expected search_key");

    c_in.searchkey.type = BincImapParserSearchKey::KEY_AND;
    c_in.searchkey.children.push_back(b);

    while (1) {
      if (!expectSPACE())
	break;

      BincImapParserSearchKey c;
      if (!expectSearchKey(c))
	throw SyntaxException("expected search_key after search_key SPACE");
      
      c_in.searchkey.children.push_back(c);
    }

    if (!expectCRLF())
      throw SyntaxException("expected CRLF after search_key");

    c_in.setName("SEARCH");
    return true;
  } catch (InputException &e) {
    e.addReason("in SearchOperator::parse");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in SearchOperator::parse");
    throw e;
  }
  return false;
}

//----------------------------------------------------------------------
bool SearchOperator::expectSearchKey(BincImapParserSearchKey &s_in) const
{
  try {
    s_in.type = BincImapParserSearchKey::KEY_OTHER;
    if (expectThisString("ALL")) s_in.name = "ALL";
    else if (expectThisString("ANSWERED")) s_in.name = "ANSWERED";
    else if (expectThisString("BCC")) {
      s_in.name = "BCC";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("BEFORE")) {
      s_in.name = "BEFORE";
      
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectDate(s_in.date))
	throw SyntaxException("expected date");
    } else if (expectThisString("BODY")) {
      s_in.name = "BODY";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("CC")) {
      s_in.name = "CC";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("DELETED")) s_in.name = "DELETED";
    else if (expectThisString("FLAGGED")) s_in.name = "FLAGGED";
    else if (expectThisString("FROM")) {
      s_in.name = "FROM";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("KEYWORD")) {
      s_in.name = "KEYWORD";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAtom(s_in.astring))
	throw SyntaxException("expected flag_keyword");
    } else if (expectThisString("NEW")) s_in.name = "NEW";
    else if (expectThisString("OLD")) s_in.name = "OLD";
    else if (expectThisString("ON")) {
      s_in.name = "ON";
      
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectDate(s_in.date))
	throw SyntaxException("expected date");      
    } else if (expectThisString("RECENT")) s_in.name = "RECENT";
    else if (expectThisString("SEEN")) s_in.name = "SEEN";
    else if (expectThisString("SINCE")) {
      s_in.name = "SINCE";
      
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectDate(s_in.date))
	throw SyntaxException("expected date");      
    } else if (expectThisString("SUBJECT")) {
      s_in.name = "SUBJECT";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("TEXT")) {
      s_in.name = "TEXT";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("TO")) {
      s_in.name = "TO";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("UNANSWERED")) s_in.name = "UNANSWERED";
    else if (expectThisString("UNDELETED")) s_in.name = "UNDELETED";
    else if (expectThisString("UNFLAGGED")) s_in.name = "UNFLAGGED";
    else if (expectThisString("UNKEYWORD")) {
      s_in.name = "UNKEYWORD";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      if (!expectAtom(s_in.astring))
	throw SyntaxException("expected flag_keyword");
    } else if (expectThisString("UNSEEN")) s_in.name = "UNSEEN";
    else if (expectThisString("DRAFT")) s_in.name = "DRAFT";
    else if (expectThisString("HEADER")) { 
      s_in.name = "HEADER";
	
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectAstring(s_in.astring))
	throw SyntaxException("expected astring");

      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectAstring(s_in.bstring))
	throw SyntaxException("expected astring");
    } else if (expectThisString("LARGER")) {
      s_in.name = "LARGER";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectNumber(s_in.number))
	throw SyntaxException("expected number");
    } else if (expectThisString("NOT")) {
      s_in.name = "NOT";
      s_in.type = BincImapParserSearchKey::KEY_NOT;
	
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      BincImapParserSearchKey s;
      if (!expectSearchKey(s))
	throw SyntaxException("expected search_key");
      s_in.children.push_back(s);
    } else if (expectThisString("OR")) {
      s_in.name = "OR";
      s_in.type = BincImapParserSearchKey::KEY_OR;
	
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      BincImapParserSearchKey s;
      if (!expectSearchKey(s))
	throw SyntaxException("expected search_key");
      s_in.children.push_back(s);

      if (!expectSPACE())
	throw SyntaxException("expected SPACE");

      BincImapParserSearchKey t;
      if (!expectSearchKey(t))
	throw SyntaxException("expected search_key");
      s_in.children.push_back(t);
    } else if (expectThisString("SENTBEFORE")) {
      s_in.name = "SENTBEFORE";
	
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectDate(s_in.date))
	throw SyntaxException("expected date");      
    } else if (expectThisString("SENTON")) {
      s_in.name = "SENTON";
	
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectDate(s_in.date))
	throw SyntaxException("expected date");      
    } else if (expectThisString("SENTSINCE")) {
      s_in.name = "SENTSINCE";
	
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectDate(s_in.date))
	throw SyntaxException("expected date");      
    } else if (expectThisString("SMALLER")) {
      s_in.name = "SMALLER";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectNumber(s_in.number))
	throw SyntaxException("expected number");
    } else if (expectThisString("UID")) {
      s_in.name = "UID";
      if (!expectSPACE())
	throw SyntaxException("expected SPACE");
	
      if (!expectSet(s_in.bset))
	throw SyntaxException("expected number");
    } else if (expectThisString("UNDRAFT")) s_in.name = "UNDRAFT";
    else if (expectSet(s_in.bset)) {
      s_in.name = "";
      s_in.type = BincImapParserSearchKey::KEY_SET;
    } else if (expectThisString("(")) {
      s_in.type = BincImapParserSearchKey::KEY_AND;

      while (1) {
	BincImapParserSearchKey c;
	if (!expectSearchKey(c))
	  throw SyntaxException("expected search_key");
	 
	s_in.children.push_back(c);
	  
	if (!expectSPACE())
	  break;
      }

      if (!expectThisString(")"))
	throw SyntaxException("expected )");
    } else
      return false;
    
    return true;

  } catch (InputException &e) {
    e.addReason("in SearchOperator::expectSearchKey");
    throw e;
  } catch (SyntaxException &e) {
    e.addReason("in SearchOperator::expectSearchKey");
    throw e;
  }
}
