/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/operator-status.cc
 *  
 *  Description:
 *    Implementation of the STATUS command.
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <iostream>

#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "io/io.h"
#include "mailbox/mailbox.h"
#include "mailbox/status.h"
#include "parsers/storage/storage.h"
#include "util/convert.h"

#include "parsers/imap/recursivedescent/recursivedescent.h"

#include "session.h"
#include "depot.h"
#include "operators.h"

using namespace ::std;
using namespace Binc;

//----------------------------------------------------------------------
StatusOperator::StatusOperator(void)
{
}

//----------------------------------------------------------------------
StatusOperator::~StatusOperator(void)
{
}

//----------------------------------------------------------------------
const string StatusOperator::getName(void) const
{
  return "STATUS";
}

//----------------------------------------------------------------------
int StatusOperator::getState(void) const
{
  return Session::AUTHENTICATED | Session::SELECTED;
}

//------------------------------------------------------------------------
void StatusOperator::process(Depot &depot, Command &command) const
{
  IO &com = IOFactory::getInstance().get(1);
  IO &logger = IOFactory::getInstance().get(2);

  try {
    Status &status = depot.getStatus(command.getMailbox());

    com << "* STATUS " << toImapString(command.getMailbox()) << " (";
    
    string prefix;
    for (vector<string>::const_iterator i = command.statuses.begin();
	 i != command.statuses.end(); ++i) {
      string tmp = *i;
      uppercase(tmp);
      if (tmp == "UIDNEXT") {
	com << prefix << "UIDNEXT " << status.getUidNext(); prefix = " ";
      } else if (tmp == "MESSAGES") {
	com << prefix << "MESSAGES " << status.getMessages(); prefix = " ";
      } else if (tmp == "RECENT") {
	com << prefix << "RECENT " << status.getRecent(); prefix = " ";
      } else if (tmp == "UIDVALIDITY") {
	com << prefix << "UIDVALIDITY " << status.getUidValidity(); prefix = " ";
      } else if (tmp == "UNSEEN") {
	com << prefix << "UNSEEN " << status.getUnseen(); prefix = " ";
      }
    }
    com << ")" << endl;
  } catch (StorageException &e) {
    logger << "Unable to status " << command.getMailbox() << ": " << e.getReason() << endl;
    throw OperatorException(Operator::NO, e.getReason());
  }
}

//----------------------------------------------------------------------
bool StatusOperator::parse(Command &c_in) const
{
  try {
    if (!expectThisString("STATUS"))
      return false;

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    string mailbox;
    if (!expectMailbox(mailbox))
      throw SyntaxException("expected mailbox");
    c_in.setMailbox(mailbox);

    if (!expectSPACE())
      throw SyntaxException("expected SPACE");

    if (!expectThisString("("))
      throw SyntaxException("expected (");

    while (1) {
      if (expectThisString("MESSAGES")) c_in.getStatuses().push_back("MESSAGES");
      else if (expectThisString("RECENT")) c_in.getStatuses().push_back("RECENT");
      else if (expectThisString("UIDNEXT")) c_in.getStatuses().push_back("UIDNEXT");
      else if (expectThisString("UIDVALIDITY")) c_in.getStatuses().push_back("UIDVALIDITY");
      else if (expectThisString("UNSEEN")) c_in.getStatuses().push_back("UNSEEN");
      else throw SyntaxException("expected status_att");

      if (!expectSPACE())
	break;
    }

    if (!expectThisString(")"))
      throw SyntaxException("expected )");
    if (!expectCRLF())
      throw SyntaxException("expected CRLF");
  } catch (InputException &e) {
    e.addReason("in OperatorStatus::parse");
    throw e;
  }

  c_in.setName("STATUS");
  return true;
}
