/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/parsers/mime/mime-parsefull.cc
 *  
 *  Description:
 *    Implementation of main mime parser components
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "mime.h"
#include "mime-utils.h"
#include "../../util/convert.h"
#include "../../io/io.h"
#include <string>
#include <vector>
#include <map>
#include <exception>
#include <iostream>

#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <errno.h>

using namespace ::std;

//------------------------------------------------------------------------
void Binc::MimeDocument::parseFull(FILE *fp) const
{
  if (allIsParsed)
    return;

  allIsParsed = true;

  while (CRLFbuffer.size() != 0)
    CRLFbuffer.pop();
  CRLFlastchar = '\0';
  CRLFbuffersize = 0;

  headerstartoffset = 0;
  headerlength = 0;
  bodystartoffset = 0;
  bodylength = 0;
  size = -1;
  messagerfc822 = false;
  multipart = false;

  int crlfnr = 0;
  int bytenr = 0;
  MimePart::parseFull(fp, "", bytenr, crlfnr);

  size = crlfnr;
}

//------------------------------------------------------------------------
int Binc::MimePart::parseFull(FILE *fp, const string &toboundary, int &bytenr, int &crlfnr) const
{
  static int level = 0;

  string name;
  string content;
  char cqueue[4];
  memset(cqueue, 0, sizeof(cqueue));

  headerstartoffset = bytenr;
  int headerstartoffsetcrlf = crlfnr;

  bool breakheader = false;
  bool quit = false;
  int c = '\0';

  while (!quit) {
    // read name
    while (1) {
      if ((c = fgetcCRLF(fp)) == EOF) {
	quit = true;
	break;
      }

      ++crlfnr;     
      if (c == '\n') ++nlines;
      if (c != '\r') ++bytenr;
      if (c == ':') break;
      if (c == '\n') {
	--bytenr;
	for (int i = name.length() - 1; i >= 0; --i) {
	  fungetcCRLF(name[i]);
	  --crlfnr;
	  if (name[i] != '\r') --bytenr;
	}
	quit = true;
	breakheader = true;
	name = "";
	break;
      }

      if (c == EOF) {
	quit = true;
	break;
      }

      name += c;

      if (name.length() == 2 && name.substr(0, 2) == "\r\n") {
	name = "";
	quit = true;
	break;
      }
    }

    if (name.length() == 1 && name[0] == '\r') {
      name = "";
      break;
    }

    if (quit) break;

    while (!quit) {
      c = fgetcCRLF(fp);
      if (c == EOF) {
	quit = true;
	break;
      }

      ++crlfnr;
      if (c == '\n') ++nlines;
      if (c != '\r') ++bytenr;

      for (int i = 0; i < 3; ++i)
	cqueue[i] = cqueue[i + 1];
      cqueue[3] = c;

      if (strncmp(cqueue, "\r\n\r\n", 4) == 0) {
	quit = true;
	break;
      }

      if (cqueue[2] == '\n') {

	// guess the mime rfc says what can not appear on the beginning
	// of a line.
	if (!isspace(cqueue[3])) {
	  if (content.length() > 2)
	    content.resize(content.length() - 2);

	  trim(content);
	  h.add(name, content);
	  
	  name = c;
	  content = "";
	  break;
	}
      }

      content += c;
    }
  }
  if (c == EOF)
    return 1;

  if (name != "") {
    if (content.length() > 2)
      content.resize(content.length() - 2);
    h.add(name, content);
  }
  
  HeaderItem ctype;
  if (h.getFirstHeader("content-type", ctype)) {
    vector<string> types;
    split(ctype.getValue(), "; \t\r\n", types);

    if (types.size() > 0) {
      // first element should describe content type
      string tmp = types[0];
      trim(tmp);
      vector<string> v;
      split(tmp, "/", v);
      string key, value;

      key = (v.size() > 0) ? v[0] : "text";
      value = (v.size() > 1) ? v[1] : "plain";
      lowercase(key);
      
      if (key == "multipart") {
	multipart = true;
	lowercase(value);
	subtype = value;
      } else if (key == "message") {
	lowercase(value);
	if (value == "rfc822")
	  messagerfc822 = true;
      }
    }

    for (vector<string>::const_iterator i = types.begin();
	 i != types.end(); ++i) {
      if ((*i).find("=") != string::npos) {
	string pair = *i;
	string key = pair.substr(0, pair.find("="));
	string value = pair.substr(pair.find("=") + 1);
	
	lowercase(key);
	string valuecp = value;
	lowercase(valuecp);
	if (key == "boundary") {
	  trim(value, "\"");
	  boundary = value;
	}
      }
    }
  }

  bool foundendofpart = false;
  headerlength = crlfnr - headerstartoffsetcrlf;
  if (headerlength < 0) headerlength = 0;

  int bodystartoffsetcrlf = crlfnr;
  bodystartoffset = bytenr;

  if (messagerfc822) {
    // message rfc822 means a completely enclosed mime document. we
    // call the parser recursively, and pass on the boundary string
    // that we got. when parse() finds this boundary, it returns.
    MimePart m;
    level += 3;
    int n = m.parseFull(fp, toboundary, bytenr, crlfnr);
    
    nbodylines += m.getNofLines();

    members.push_back(m);
    level -= 3;

    // if the enclosed doc ended with _our_ boundary string
    if (n == 1)
      foundendofpart = true;
    else
      foundendofpart = false;

  } else if (multipart) {
    // multipart parsing starts with skipping to the first
    // boundary. then we call parse() for all parts. the last parse()
    // command will return a code indicating that it found the last
    // boundary of this multipart.
    string _boundary = "--"; _boundary += boundary;

    char *boundaryqueue = NULL;
    int endpos = _boundary.length();
    boundaryqueue = new char[endpos];
    int boundarypos = 0;
    bool eof = false;

    // first, skip to the first _boundary string
    while (1) {
      if ((c = fgetcCRLF(fp)) == EOF) {
	eof = true;
	break;
      }
      ++crlfnr;
      if (c == '\n') { ++nbodylines; ++nlines; }
      if (c != '\r') ++bytenr;

      boundaryqueue[boundarypos++ % endpos] = c;
      
      if (compareStringToQueue(_boundary, boundaryqueue, boundarypos, endpos)) {
	boundaryqueue[0] = '\0';
	
	int a = fgetcCRLF(fp); 

	++crlfnr;
	if (a == '\n') { /* ++nbodylines; ++nlines; */ }
	if (a != '\r') ++bytenr;
	if (a == EOF) {
	  quit = true;
	  break;
	}

	int b = fgetcCRLF(fp); 

	++crlfnr;
	if (b == '\n') { /* ++nbodylines; ++nlines;*/ }
	if (b != '\r') ++bytenr;
	if (b == EOF) {
	  quit = true;
	  break;
	}

	if (a == '-' && b == '-') {
	  int a = fgetcCRLF(fp);
	  ++crlfnr;
	  if (b == '\n') { /* ++nbodylines; ++nlines; */ }
	  if (a != '\r') ++bytenr;

	  if (a == EOF) {
	    quit = true;
	    break;
	  }
	  
	  if (a == '\r') {
	    int b = fgetcCRLF(fp);
	    ++crlfnr;
	    if (b == '\n') { /* ++nbodylines; ++nlines; */ }
	    if (b != '\r') ++bytenr;

	    if (b != '\n') {
	      // FIXME: barf!
	    }
	  } else {
	    // FIXME: barf!!
	  }
	  
	  foundendofpart = true;
	}
	
	break;
      }
    }

    if (boundaryqueue != NULL)
      delete boundaryqueue;

    // at this point, we found the first boundary and we ate it. the next
    // consecutive calls to parse() will all have given that the start of
    // the input is the start of the part, straight after the boundary.
    if (!eof && !foundendofpart) {
      // then, parse all the parts
      bool quit = false;
      while (!quit) {
	MimePart m;

	level += 3;
	if (m.parseFull(fp, boundary, bytenr, crlfnr) == 1)
	  quit = true;

	members.push_back(m);
	level -= 3;
      }
    }

    // we mustn't forget to eat our own parent's boundary, if it's non-empty.
    if (!eof && toboundary != "") {
      string _toboundary;

      if (toboundary != "") {
	_toboundary = "--";
	_toboundary += toboundary;
      }
      
      char *boundaryqueue = NULL;
      int endpos = _toboundary.length();
      if (_toboundary != "")
	boundaryqueue = new char[endpos];
      int boundarypos = 0;
      
      string line;
      int nchars = 0;
      while ((c = fgetcCRLF(fp)) != EOF) {
	++crlfnr;
	if (c != '\r') ++bytenr;
	if (c == '\n') { ++nbodylines; ++nlines; }
	nchars++;
	
	if (toboundary != "") {
	  // find boundary
	  boundaryqueue[boundarypos++ % endpos] = c;
	  
	  if (compareStringToQueue(_toboundary, boundaryqueue, boundarypos, endpos)) {
	    boundaryqueue[0] = '\0';
	    int a = fgetcCRLF(fp);
	    ++crlfnr;
	    if (a == '\n') { /* ++nbodylines; ++nlines; */ }
	    if (a != '\r') ++bytenr;
	    if (a == EOF) {
	      quit = true;
	      break;
	    }
	    int b = fgetcCRLF(fp);
	    ++crlfnr;
	    if (b == '\n') { /* ++nbodylines; ++nlines; */ }
	    if (b != '\r') ++bytenr;
	    if (b == EOF) {
	      quit = true;
	      break;
	    }

	    if (a == '-' && b == '-') {
	      int a = fgetcCRLF(fp);
	      ++crlfnr;
	      if (a == '\n') { /* ++nbodylines; ++nlines; */ }
	      if (a != '\r') ++bytenr;
	      if (a == EOF) {
		quit = true;
		break;
	      }

	      if (a == '\r') {
		int b = fgetcCRLF(fp);
		++crlfnr;
		if (b == '\n') { /* ++nbodylines; ++nlines; */ }
		if (b != '\r') ++bytenr;
		if (b != '\n') {
		  // FIXME: barf!
		}
	      } else {
		// FIXME: barf!!
	      }
	      
	      foundendofpart = true;
	    }

	    break;
	  }
	}
      }

      if (boundaryqueue != NULL)
	delete boundaryqueue;
    }

  } else {
    string _toboundary;

    if (toboundary != "") {
      _toboundary = "--";
      _toboundary += toboundary;
    }

    char *boundaryqueue = NULL;
    int endpos = _toboundary.length();
    if (_toboundary != "")
      boundaryqueue = new char[endpos];
    int boundarypos = 0;

    string line;
    int nchars = 0;
    while ((c = fgetcCRLF(fp)) != EOF) {
      crlfnr++;
      if (c != '\r') ++bytenr;
      if (c == '\n') { ++nbodylines; ++nlines; }
      nchars++;
      
      if (toboundary != "") {
	// find boundary
	boundaryqueue[boundarypos++ % endpos] = c;

	if (compareStringToQueue(_toboundary, boundaryqueue, boundarypos, endpos)) {
	  boundaryqueue[0] = '\0';
	  int a = fgetcCRLF(fp);
	  ++crlfnr;
	  if (a == '\n') { /* ++nbodylines; ++nlines; */ }
	  if (a != '\r') ++bytenr;
	  if (a == EOF) {
	    quit = true;
	    break;
	  }
	  int b = fgetcCRLF(fp); 
	  ++crlfnr;
	  if (b == '\n') { /* ++nbodylines; ++nlines; */ }
	  if (b != '\r') ++bytenr;
	  if (b == EOF) {
	    quit = true;
	    break;
	  }

	  if (a == '-' && b == '-') {
	    int a = fgetcCRLF(fp); 
	    ++crlfnr;
	    if (a == '\n') { /* ++nbodylines; ++nlines; */ }
	    if (a != '\r') ++bytenr;
	    if (a == EOF) {
	      quit = true;
	      break;
	    }

	    if (a == '\r') {
	      int b = fgetcCRLF(fp);
	      ++crlfnr;
	      if (b == '\n') { /* ++nbodylines; ++nlines; */ }
	      if (b != '\r') ++bytenr;
	      if (b != '\n') {
		// FIXME: barf!
	      }
	    } else {
	      // FIXME: barf!!
	    }

	    foundendofpart = true;
	  }

	  break;
	}
      }
    }

    if (boundaryqueue != NULL)
      delete boundaryqueue;
  }

  bodylength = crlfnr - bodystartoffsetcrlf;

  // FIXME: is this crap?
  if (toboundary != "") {
    bodylength -= toboundary.length() + 4;
    if (foundendofpart)
      bodylength -= 4;
  }

  if (bodylength < 0) bodylength = 0;

  if (foundendofpart)
    return 1;
  else
    return 0;
}
