/* -*- Mode: c++; -*- */
/*  --------------------------------------------------------------------
 *  Filename:
 *    src/session.cc
 *  
 *  Description:
 *    <--->
 *
 *  Authors:
 *    Andreas Aardal Hanssen <bincimap@andreas.hanssen.name>
 *
 *  Bugs:
 *
 *  ChangeLog:
 *
 *  --------------------------------------------------------------------
 *  Copyright 2002, 2003 Andreas Aardal Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *  --------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <syslog.h>

#include "io/io.h"
#include "session.h"
#include "util/tools.h"
#include "util/convert.h"
#include <string>
#include <map>

using namespace ::std;
using namespace Binc;
using namespace ArgParser;

extern char **environ;

//----------------------------------------------------------------------
Session::Session(void)
{
  readbytes = 0;
  writebytes = 0;
  statements = 0;
  bodies = 0;
  logfacility = LOG_DAEMON;
}

//----------------------------------------------------------------------
Session &Session::getInstance(void)
{
  static Session session;
  return session;
}

//----------------------------------------------------------------------
void Session::initConfig(void)
{
  args.assign();
}

//----------------------------------------------------------------------
const int Session::getState(void) const
{
  return state;
}

//----------------------------------------------------------------------
void Session::setState(int n)
{
  state = n;
}

//----------------------------------------------------------------------
const string &Session::getUserID(void) const
{
  return userid;
}

//----------------------------------------------------------------------
void Session::setUserID(const string &s)
{
  userid = s;
}

//----------------------------------------------------------------------
const string &Session::getIP(void) const
{
  return ip;
}

//----------------------------------------------------------------------
void Session::setIP(const string &s)
{
  ip = s;
}

//----------------------------------------------------------------------
void Session::setLogFacility(int facility)
{
  logfacility = facility;
}

//----------------------------------------------------------------------
int Session::getLogFacility(void) const
{
  return logfacility;
}

//----------------------------------------------------------------------
void Session::addBody(void)
{
  ++bodies;
}

//----------------------------------------------------------------------
void Session::addStatement(void)
{
  ++statements;
}

//----------------------------------------------------------------------
void Session::addReadBytes(int i)
{
  readbytes += i;
}

//----------------------------------------------------------------------
void Session::addWriteBytes(int i)
{
  writebytes += i;
}

//----------------------------------------------------------------------
int Session::getBodies(void) const
{
  return bodies;
}

//----------------------------------------------------------------------
int Session::getStatements(void) const
{
  return statements;
}

//----------------------------------------------------------------------
int Session::getWriteBytes(void) const
{
  return writebytes;
}

//----------------------------------------------------------------------
int Session::getReadBytes(void) const
{
  return readbytes;
}

void Session::parseCommandLine(int argc, char * argv[])
{
  args = ArgParser::Args(argc, argv);

  args.addOptional("h|?|help", command.help);
  args.addOptional("v|version", command.version);
  args.addOptional("s|ssl", command.ssl);
  args.addOptional("c|conf", command.configfile);

  // Override config file settings with command line options
  string arg;
  globalconfig.setSection("Authentication");

  args.addOptional("a|allow-plain", 
		   globalconfig.get("allow plain auth in non ssl"));
  args.addOptional("p|auth-penalty",
		   globalconfig.get("auth penalty"));
  args.addOptional("d|disable-starttls",
		   globalconfig.get("disable starttls"));

  globalconfig.setSection("Log");

  args.addOptional("L|logtype", globalconfig.get("type"));
  args.addOptional("I|ip-variable", globalconfig.get("environment ip variable"));

  globalconfig.setSection("Mailbox");

  args.addOptional("M|mailbox-type", globalconfig.get("type"));
  args.addOptional("m|mailbox-path", globalconfig.get("path"));
  args.addOptional("C|create-inbox", globalconfig.get("auto create inbox"));
  args.addOptional("S|subscribe-mailboxes",
		   globalconfig.get("auto subscribe mailboxes"));
  args.addOptional("u|umask", globalconfig.get("umask"));

  globalconfig.setSection("Security");

  args.addOptional("J|jail-path", globalconfig.get("jail path"));
  args.addOptional("x|jail-user", globalconfig.get("jail user"));
  args.addOptional("X|jail-group", globalconfig.get("jail group"));

  globalconfig.setSection("Session");

  args.addOptional("i|idle-timeout", globalconfig.get("idle timeout"));
  args.addOptional("t|auth-timeout", globalconfig.get("auth timeout"));
  args.addOptional("T|transfer-timeout", globalconfig.get("transfer timeout"));
  args.addOptional("b|transfer-buffersize", 
		   globalconfig.get("transfer buffer size"));

  globalconfig.setSection("SSL");

  args.addOptional("p|pem-file", globalconfig.get("pem file"));
  args.addOptional("P|ca-path", globalconfig.get("ca path"));
  args.addOptional("f|ca-file", globalconfig.get("ca file"));
  args.addOptional("l|cipher-list", globalconfig.get("cipher list"));
  args.addOptional("V|verify-peer", globalconfig.get("verify peer"));

  try {
    int i = args.parse(restArgs);
    unparsedArgs = argv + i;
  }
  catch (const MissingArgumentException & e) {
    throw SyntaxException(e.name() + " requires a value");
  }
  catch (const UnknownArgumentException & e) {
    throw SyntaxException(e.name() + " is not recognized as a valid argument");
  }
  catch (const RequiredArgumentException & e) {
    set<string>::const_iterator i = e.missing.begin();
    string tmp;
    while (i != e.missing.end()) {
      tmp += " " + *i;
      ++i;
    }
    throw SyntaxException("Missing required argument(s):" + tmp);

  }
  catch (const RequiredCommandException & e) {
    throw SyntaxException("The argument '" + e.name() + "' is required.");
  }
  catch (const CheckException & e) {
    throw SyntaxException("Invalid value for parameter '" +e.name() + "'");
  }
  catch (...) {
    SyntaxException("Failed to parse command line arguments.");
  }
  args.assign("conf");  
  args.assign("help");
  args.assign("version");
}

//----------------------------------------------------------------------
void Session::add(const string &a, const string &b)
{
  attrs[a] = b;
}

//----------------------------------------------------------------------
const string &Session::operator [] (const string &s_in) const
{
  static const string NIL = "";
  if (attrs.find(s_in) == attrs.end())
    return NIL;
  else
    return attrs.find(s_in)->second;
}

//----------------------------------------------------------------------
void Session::exportToEnv(void) throw (BincException &)
{
  IO &com = IOFactory::getInstance().get(1);
  Tools &tools = Tools::getInstance();

  tools.setenv("BINCIMAP_STATE", toString(state));
  tools.setenv("BINCIMAP_USERID", userid);
  tools.setenv("BINCIMAP_IP", ip);

  for (map<string, string>::const_iterator i = attrs.begin();
       i != attrs.end(); ++i)
    tools.setenv("BINCIMAP_CONF_" + i->first, i->second);

  int n = 0;
  for (vector<string>::const_iterator i = subscribed.begin();
       i != subscribed.end(); ++i, ++n)
    tools.setenv("BINCIMAP_SUBSCRIBED_" + toString(n), *i);

  string section;
  if (globalconfig.getFirstSection(section))
    do {
      string key, value;

      globalconfig.setSection(section);
      if (globalconfig.getFirstEntry(key, value))
	do {
	  tools.setenv("BINCIMAP_GLOBALCONFIG_" + toHex(section) + "::" + toHex(key), toHex(value));
	} while (globalconfig.getNextEntry(key, value));
    } while (globalconfig.getNextSection(section));

  if (localconfig.getFirstSection(section))
    do {
      string key, value;
      localconfig.setSection(section);
      if (localconfig.getFirstEntry(key, value))
	do {
	  tools.setenv("BINCIMAP_LOCALCONFIG_" + toHex(section) + "::" + toHex(key), toHex(value));
	} while (localconfig.getNextEntry(key, value));
    } while (localconfig.getNextSection(section));
}

//----------------------------------------------------------------------
void Session::importFromEnv(void) throw (BincException &)
{
  IO & com = IOFactory::getInstance().get(1);
  Tools &tools = Tools::getInstance();

  char *c;
  int cnt = 0;
  while ((c = environ[cnt]) != NULL) {
    string s = c;

    if (s.substr(0, 14) == "BINCIMAP_STATE") state = atoi(s.substr(15));
    else if (s.substr(0, 15) == "BINCIMAP_USERID") userid = s.substr(16);
    else if (s.substr(0, 11) == "BINCIMAP_IP") ip = s.substr(12);
    else if (s.substr(0, 21) == "BINCIMAP_GLOBALCONFIG") {
      const string config = s.substr(22);
      if (config.find("::") != string::npos) {
	const string section = fromHex(config.substr(0, config.find("::")));
	const string data = config.substr(config.find("::") + 2);
	
	if (data.find("=") != string::npos) {
	  const string key = fromHex(data.substr(0, data.find("=")));
	  const string value = fromHex(data.substr(data.find("=") + 1));

	  globalconfig.setSection(section);
	  globalconfig.insert(key, value);
	}
      }
    } else if (s.substr(0, 20) == "BINCIMAP_LOCALCONFIG") {
      const string config = s.substr(21);
      if (config.find("::") != string::npos) {
	const string section = fromHex(config.substr(0, config.find("::")));
	const string data = config.substr(config.find("::") + 2);
	
	if (data.find("=") != string::npos) {
	  const string key = fromHex(data.substr(0, data.find("=")));
	  const string value = fromHex(data.substr(data.find("=") + 1));

	  localconfig.setSection(section);
	  localconfig.insert(key, value);
	}
      }
    }

    ++cnt;
  }
}
